/* 
 * File:   Exceptions.h
 * Author: lubajz
 *
 * Created on 12. prosinec 2009, 21:37
 */

#ifndef _EXCEPTIONS_H
#define	_EXCEPTIONS_H

class GeneralException {
public:
    char * message;
    GeneralException(char * _message);
    void displayMessage();
};

class SumProbabilitiesException : public GeneralException {
public:
    double sum;
    SumProbabilitiesException(double _sum);
    void displayMessage();
};

class NumParametersException : public GeneralException {
public:
    NumParametersException();
};

class FileNotExists : public GeneralException {
public:
    char * fileName;
    FileNotExists(char * _fileName);
    void displayMessage();
};

class NoOccurenceOfVariableInLSTFile : public GeneralException {
public:
    char * nameOfVariable;
    char * fileName;
    NoOccurenceOfVariableInLSTFile(char * _nameOfVariable, char * _fileName);
    void displayMessage();
};



#endif	/* _EXCEPTIONS_H */

